use std::collections::HashSet;

use glutin::{
    event::{
        ElementState, Event, KeyboardInput, MouseScrollDelta, WindowEvent as GlutinWindowEvent,
    },
    event_loop::{ControlFlow, EventLoop},
    platform::desktop::EventLoopExtDesktop,
};
use shrev::EventChannel;
use specs::World;

pub use glutin::event::{MouseButton, VirtualKeyCode};

pub struct Events {
    keys: HashSet<VirtualKeyCode>,
    event_loop: EventLoop<()>,
}

impl Events {
    pub fn new(world: &mut World) -> Self {
        world.insert(EventChannel::<MouseEvent>::default());
        world.insert(EventChannel::<WindowEvent>::default());
        world.insert(EventChannel::<KeyboardEvent>::default());

        Self {
            keys: HashSet::new(),
            event_loop: EventLoop::new(),
        }
    }
}

impl Events {
    pub fn handle(&self) -> &EventLoop<()> {
        &self.event_loop
    }

    pub fn process(&mut self, world: &World) {
        let mut mouse_events = world.fetch_mut::<EventChannel<MouseEvent>>();
        let mut window_events = world.fetch_mut::<EventChannel<WindowEvent>>();
        let mut keyboard_events = world.fetch_mut::<EventChannel<KeyboardEvent>>();

        let keys = &mut self.keys;
        let event_loop = &mut self.event_loop;

        event_loop.run_return(|event, _target, flow_control| {
            *flow_control = ControlFlow::Poll;

            match event {
                Event::MainEventsCleared | Event::RedrawRequested(_) => {
                    *flow_control = ControlFlow::Exit;
                }
                Event::WindowEvent { event, .. } => match event {
                    /* Mouse Events */
                    GlutinWindowEvent::CursorMoved { position, .. } => {
                        mouse_events
                            .single_write(MouseEvent::Move(position.x as _, position.y as _));
                    }
                    GlutinWindowEvent::MouseInput {
                        state: ElementState::Pressed,
                        button,
                        ..
                    } => {
                        mouse_events.single_write(MouseEvent::ButtonDown(button));
                    }
                    GlutinWindowEvent::MouseInput {
                        state: ElementState::Released,
                        button,
                        ..
                    } => {
                        mouse_events.single_write(MouseEvent::ButtonUp(button));
                    }
                    GlutinWindowEvent::MouseWheel {
                        delta: MouseScrollDelta::LineDelta(x, y),
                        ..
                    } => {
                        if x != 0.0 {
                            mouse_events.single_write(MouseEvent::ScrollX(x));
                        }

                        if y != 0.0 {
                            mouse_events.single_write(MouseEvent::ScrollY(y));
                        }
                    }

                    /* Key Event */
                    GlutinWindowEvent::KeyboardInput {
                        input:
                            KeyboardInput {
                                state: ElementState::Pressed,
                                virtual_keycode: Some(key),
                                ..
                            },
                        ..
                    } => {
                        if keys.insert(key) {
                            keyboard_events.single_write(KeyboardEvent::KeyDown(key));
                        }

                        keyboard_events.single_write(KeyboardEvent::KeyPress(key));
                    }
                    GlutinWindowEvent::KeyboardInput {
                        input:
                            KeyboardInput {
                                state: ElementState::Released,
                                virtual_keycode: Some(key),
                                ..
                            },
                        ..
                    } => {
                        if keys.remove(&key) {
                            keyboard_events.single_write(KeyboardEvent::KeyUp(key));
                        }
                    }

                    /* Window Event */
                    GlutinWindowEvent::Resized(pos) => {
                        window_events.single_write(WindowEvent::Resize(pos.width, pos.height));
                    }
                    GlutinWindowEvent::CloseRequested => {
                        window_events.single_write(WindowEvent::Close);
                    }

                    /* Ignore */
                    _ => (),
                },
                _ => (),
            }
        });
    }
}

pub enum WindowEvent {
    Resize(u32, u32),
    Close,
}

pub enum MouseEvent {
    Move(f32, f32),
    ButtonDown(MouseButton),
    ButtonUp(MouseButton),
    ScrollX(f32),
    ScrollY(f32),
}

pub enum KeyboardEvent {
    KeyUp(VirtualKeyCode),
    KeyDown(VirtualKeyCode),
    KeyPress(VirtualKeyCode),
}
