use glc::{
    matrix::Matrix4f,
    misc::Bindable,
    shader::{Program, Type},
    vector::Vector3f,
};
use log::error;
use shrev::{EventChannel, ReaderId};
use space_crush_common::misc::WorldHelper as _;
use specs::{prelude::*, ReadExpect, System, World, WriteExpect};

use crate::{
    constants::{UNIFORM_BUFFER_INDEX_CAMERA, UNIFORM_BUFFER_INDEX_UNIFORM},
    misc::{MouseEvent, WorldHelper},
    resources::{Camera, Config, Geometry, State, Uniform},
    Error,
};

pub struct Init {
    program: Program,
    resolution: (u32, u32),
    mouse_event_id: ReaderId<MouseEvent>,
}

impl Init {
    pub fn new(world: &mut World) -> Result<Self, Error> {
        let program = world.load_program(vec![
            (Type::Vertex, "resources/shader/noise/vert.glsl"),
            (Type::Fragment, "resources/shader/noise/frag.glsl"),
        ])?;
        let resolution = (0, 0);
        let mouse_event_id = world.register_event_reader::<MouseEvent>()?;

        world
            .resource::<Camera>()?
            .bind(UNIFORM_BUFFER_INDEX_CAMERA)?;
        world
            .resource::<Uniform>()?
            .bind(UNIFORM_BUFFER_INDEX_UNIFORM)?;

        Ok(Self {
            program,
            resolution,
            mouse_event_id,
        })
    }
}

#[derive(SystemData)]
pub struct InitData<'a> {
    camera: WriteExpect<'a, Camera>,
    uniform: WriteExpect<'a, Uniform>,
    state: ReadExpect<'a, State>,
    config: ReadExpect<'a, Config>,
    geometry: ReadExpect<'a, Geometry>,
    mouse_events: ReadExpect<'a, EventChannel<MouseEvent>>,
}

impl<'a> System<'a> for Init {
    type SystemData = InitData<'a>;

    fn run(&mut self, data: Self::SystemData) {
        let InitData {
            mut camera,
            mut uniform,
            state,
            config,
            geometry,
            mouse_events,
        } = data;

        /* screen size */
        if self.resolution != state.resolution {
            self.resolution = state.resolution;

            gl::viewport(0, 0, self.resolution.0 as _, self.resolution.1 as _);

            if let Err(err) = camera.resize(self.resolution.0 as _, self.resolution.1 as _) {
                error!("Error while updating camera: {}", err);
            }
        }

        /* zoom */
        let events = mouse_events.read(&mut self.mouse_event_id);
        for event in events {
            if let MouseEvent::ScrollY(delta) = event {
                let s = config.input.speed_camera_zoom;
                let z = s / (s - delta);
                let m = Matrix4f::translate((state.mouse_pos.0, state.mouse_pos.1, 0.0).into())
                    * Matrix4f::scale(z.into())
                    * Matrix4f::translate((-state.mouse_pos.0, -state.mouse_pos.1, 0.0).into());

                if let Err(err) = camera.update_with(move |v| m * v) {
                    error!("Error while zooming camera: {}", err);
                }
            }
        }

        /* move camera */
        let up = state.key_state(&config.input.key_camera_move_up);
        let down = state.key_state(&config.input.key_camera_move_down);
        let left = state.key_state(&config.input.key_camera_move_left);
        let right = state.key_state(&config.input.key_camera_move_right);

        if up || down || left || right {
            let s = config.input.speed_camera_move;
            let translate = Vector3f::new(
                if left { s } else { 0.0 } + if right { -s } else { 0.0 },
                if up { -s } else { 0.0 } + if down { s } else { 0.0 },
                0.0,
            );

            let m = Matrix4f::translate(translate);
            if let Err(err) = camera.update_with(move |v| m * v) {
                error!("Error while moving camera: {}", err);
            }
        }

        if let Err(err) = uniform.update() {
            error!("Error while updating global uniform data: {}", err);
        }

        /* render background */
        self.program.bind();
        geometry.render_quad();
        self.program.unbind();
    }
}
