use glc::{
    matrix::Matrix4f,
    misc::Bindable,
    shader::{Program, Type, Uniform},
    texture::Texture,
    vector::Vector4f,
};
use log::error;
use space_crush_common::components::{Planet, Position};
use specs::{prelude::*, ReadExpect, ReadStorage, System, World};

use crate::{
    constants::{UNIFORM_BUFFER_INDEX_CAMERA, UNIFORM_BUFFER_INDEX_UNIFORM},
    misc::WorldHelper,
    resources::Geometry,
    Error,
};

pub struct Planets {
    program: Program,
    texture: Texture,
    model_location: gl::GLint,
}

impl Planets {
    pub fn new(world: &World) -> Result<Self, Error> {
        let program = world.load_program(vec![
            (Type::Vertex, "resources/shader/planet/vert.glsl"),
            (Type::Fragment, "resources/shader/planet/frag.glsl"),
        ])?;

        program.uniform_block_binding("Camera", UNIFORM_BUFFER_INDEX_CAMERA)?;
        program.uniform_block_binding("Global", UNIFORM_BUFFER_INDEX_UNIFORM)?;

        let glow_color = Vector4f::new(1.0, 1.0, 1.0, 0.1);
        let glow_color = Uniform::Vector4f(&glow_color);
        let model_location = program.uniform_location("uModel")?;

        program.bind();
        program.uniform("uTexture", Uniform::Texture(0))?;
        program.uniform("uGlowColor", glow_color)?;
        program.unbind();

        let texture = world.load_texture("resources/textures/planet01.png")?;

        Ok(Self {
            program,
            texture,
            model_location,
        })
    }
}

#[derive(SystemData)]
pub struct PlanetsData<'a> {
    geometry: ReadExpect<'a, Geometry>,
    position: ReadStorage<'a, Position>,
    planet: ReadStorage<'a, Planet>,
}

impl<'a> System<'a> for Planets {
    type SystemData = PlanetsData<'a>;

    fn run(&mut self, data: Self::SystemData) {
        let PlanetsData {
            geometry,
            position,
            planet,
        } = data;

        gl::enable(gl::BLEND);
        gl::blend_func(gl::SRC_ALPHA, gl::ONE_MINUS_SRC_ALPHA);

        self.texture.bind();
        self.program.bind();

        for (p, _) in (&position, &planet).join() {
            let x = p.pos.x;
            let y = p.pos.y;
            let s = p.size;

            let m = Matrix4f::new(
                Vector4f::new(s, 0.0, 0.0, 0.0),
                Vector4f::new(0.0, s, 0.0, 0.0),
                Vector4f::new(0.0, 0.0, s, 0.0),
                Vector4f::new(x, y, 0.0, 1.0),
            );

            if let Err(err) = self
                .program
                .uniform(self.model_location, Uniform::Matrix4f(&m))
            {
                error!("Error while updating model matrix: {}", err);
            }

            geometry.render_quad();
        }

        self.program.unbind();
        self.texture.unbind();

        gl::disable(gl::BLEND);
    }
}
